#include <sys/stat.h>

#include <gst/gst.h>

#include "Eic.h"


GstElement          *pipeline;

static void new_decoded_pad_cb (GstElement * decodebin,
                                GstPad * new_pad,
                                gboolean last,
                                GstElement * pipeline);
static void element_added_cb (GstBin *bin,
                              GstElement *element,
                              gpointer user_data);

/* the result must be freed */
char *
get_size_string (gsize st_size)
{
  double dsize;
  char   size[1024];

  dsize = (double)st_size;

  if (dsize < 1024)
    sprintf (size, "%'.0f B", dsize);
  else
    {
      dsize /= 1024.0;
      if (dsize < 1024)
	sprintf (size, "%'.1f KB (or %'.0f B)",
		 dsize, (double)st_size);
      else
	{
	  dsize /= 1024.0;
	  if (dsize < 1024)
	    sprintf (size, "%'.1f MB (or %'.1f KB, or %'.0f B)",
		     dsize, (double)st_size/1024, (double)st_size);
	  else
	    {
	      dsize /= 1024.0;
	      sprintf (size, "%'.1f GB (or %'.1f MB or %'.1f KB, or %'.0f B)",
		       dsize, (double)st_size/(1024*1024),
		       (double)st_size/1024, (double)st_size);
	    }
	}
    }

  return g_strndup (size, 1024);
}

Eic_Factory_Info *
eic_factory_new (GstElementFactory *factory)
{
  Eic_Factory_Info  *info;

  info = (Eic_Factory_Info *)g_malloc (sizeof (Eic_Factory_Info));
  if (!info)
    return NULL;

  info->name = gst_element_factory_get_longname (factory);
  info->description = gst_element_factory_get_description (factory);
  info->class = gst_element_factory_get_klass (factory);
  info->authors = g_strsplit (gst_element_factory_get_author (factory), ", ", 0);
  return info;
}

void
eic_factory_free (Eic_Factory_Info  *info)
{
  if (!info)
    return;

  if (info->authors)
    g_strfreev (info->authors);
  g_free (info);
}

void
eic_video_prop_free (Eic_Video_Prop *vp)
{
  if (!vp)
    return;

  if (vp->decoder_info)
    eic_factory_free (vp->decoder_info);
  if (vp->prop)
    g_free (vp->prop);
  g_free (vp);
}

void
eic_audio_prop_free (Eic_Audio_Prop *ap)
{
  if (!ap)
    return;

  if (ap->decoder_info)
    eic_factory_free (ap->decoder_info);
  if (ap->prop)
    g_free (ap->prop);
  g_free (ap);
}

Eic_Prop *
eic_new (const char *filename)
{
  Eic_Prop   *eic;
  struct stat buf;

  if (!filename)
    return NULL;

  eic = (Eic_Prop *)g_malloc (sizeof (Eic_Prop));
  if (!eic)
    return NULL;

  eic->filename = g_strdup (filename);
  stat (filename, &buf);
  eic->size = buf.st_size;
  eic->demuxer_info = NULL;
  eic->video_prop = NULL;
  eic->audio_prop = NULL;

  return eic;
}

void
eic_free (Eic_Prop *eic)
{
  GList *list;

  if (!eic)
    return;

  if (eic->filename)
    g_free (eic->filename);

  if (eic->demuxer_info)
    eic_factory_free (eic->demuxer_info);
  if (eic->video_prop) {
    list = g_list_first (eic->video_prop);
    for (; list ; list = g_list_next (list))
      {
        Eic_Video_Prop *vp;
        
        vp = (Eic_Video_Prop *)list->data;
        eic_video_prop_free (vp);
      }
    g_list_free (eic->video_prop);    
  }

  if (eic->audio_prop) {
    list = g_list_first (eic->audio_prop);
    for (; list ; list = g_list_next (list))
      {
        Eic_Audio_Prop *ap;
        
        ap = (Eic_Audio_Prop *)list->data;
        eic_audio_prop_free (ap);
      }
    g_list_free (eic->audio_prop);    
  }

  g_free (eic);

  gst_object_unref (pipeline);
  gst_deinit ();
}

int
eic_fill (Eic_Prop *eic)
{
  GstElement          *filesrc;
  GstElement          *decodebin;
  GstIterator         *it;
  gpointer             data;
  GstStateChangeReturn res;

  if (!eic)
    return 0;

  gst_init (NULL, NULL);

  pipeline = gst_pipeline_new ("pipeline");

  filesrc = gst_element_factory_make ("filesrc", "filesrc");
  g_assert (filesrc);
  g_object_set (G_OBJECT (filesrc), "location", eic->filename, NULL);

  decodebin = gst_element_factory_make ("decodebin", "decodebin");
  g_assert (decodebin);
  g_signal_connect (G_OBJECT (decodebin), "new-decoded-pad",
      G_CALLBACK (new_decoded_pad_cb), pipeline);
  g_signal_connect (G_OBJECT (decodebin), "element-added",
      G_CALLBACK (element_added_cb), eic);

  gst_bin_add_many (GST_BIN (pipeline), filesrc, decodebin, NULL);
  gst_element_link (filesrc, decodebin);

  res = gst_element_set_state (pipeline, GST_STATE_PAUSED);
  if (res == GST_STATE_CHANGE_FAILURE) {
    printf ("Could not go to PAUSED state\n");
    gst_object_unref (pipeline);
    return 0;
  }
  res = gst_element_get_state (pipeline, NULL, NULL, GST_CLOCK_TIME_NONE);
  if (res != GST_STATE_CHANGE_SUCCESS) {
    printf ("Failed to complete state change to PAUSED\n");
    gst_object_unref (pipeline);
    return 0;
  }

  it = gst_element_iterate_src_pads (decodebin);
  while (gst_iterator_next (it, &data) == GST_ITERATOR_OK) {
    GstPad            *pad;
    GstCaps           *caps;
    gchar             *str;
    GstQuery          *query;
    gint64             duration = -1;
    GstElement        *decoder;
    GstElementFactory *factory;
    Eic_Factory_Info  *decoder_info;
    const char        *mimetype;

    pad = GST_PAD (data);
    if (GST_IS_GHOST_PAD(pad)) {
      GstPad   *decoder_pad;
      
      decoder_pad = gst_ghost_pad_get_target (GST_GHOST_PAD (pad));
      decoder = gst_pad_get_parent_element (decoder_pad);
      gst_object_unref (decoder_pad);
    }
    else {
      decoder =  gst_pad_get_parent_element (pad);
    }
    {
      GstCaps *decoder_caps;

      decoder_caps = GST_PAD_CAPS (gst_element_get_pad(decoder, "sink"));
      mimetype = gst_structure_get_name (gst_caps_get_structure (decoder_caps, 0));
    }
    factory = gst_element_get_factory (decoder);
    decoder_info = eic_factory_new (factory);

    gst_object_unref (decoder);

    caps = gst_pad_get_caps (pad);
    str = gst_caps_to_string (caps);

    query = gst_query_new_duration (GST_FORMAT_TIME);
    if (gst_pad_query (pad, query))
      gst_query_parse_duration (query, NULL, &duration);
    gst_query_unref (query);

    if (g_str_has_prefix (str, "video/")) {
      Eic_Video_Prop *vp;
      GstStructure   *structure;
      const GValue   *val;

      vp = (Eic_Video_Prop *)g_malloc (sizeof (Eic_Video_Prop));
      if (!vp) continue;
      eic->video_prop = g_list_append (eic->video_prop, vp);
      vp->mimetype = mimetype;
      vp->decoder_info = decoder_info;
      vp->length_time = duration;
      vp->prop = NULL;

      structure = gst_caps_get_structure (GST_CAPS (caps), 0);
      gst_structure_get_int (structure, "width", &vp->width);
      gst_structure_get_int (structure, "height", &vp->height);
      val = gst_structure_get_value (structure, "framerate");
      vp->framerate_num = gst_value_get_fraction_numerator (val);
      vp->framerate_den = gst_value_get_fraction_denominator (val);
      val = gst_structure_get_value (structure, "pixel-aspect-ratio");
      if (val) {
        vp->par_num = gst_value_get_fraction_numerator (val);
        vp->par_den = gst_value_get_fraction_denominator (val);
      }
      else{
        vp->par_num = 1;
        vp->par_den = 1;
      }
      val = g_malloc0 (sizeof (GValue));
      val = g_value_init ((GValue *)val, GST_TYPE_FRACTION);
      gst_value_set_fraction ((GValue *)val, vp->width * vp->par_num, vp->height * vp->par_den);
      vp->dar_num = gst_value_get_fraction_numerator (val);
      vp->dar_den = gst_value_get_fraction_denominator (val);
      g_free ((GValue *)val);
      vp->length_frames = vp->length_time * vp->framerate_num / ( vp->framerate_den * GST_SECOND);

      if (g_str_has_prefix (str, "video/x-raw-yuv")) {
        Eic_Video_Yuv_Prop *vyp;

        vyp = (Eic_Video_Yuv_Prop *)g_malloc (sizeof (Eic_Video_Yuv_Prop));
        if (vyp) {
          guint32 fourcc;

          gst_structure_get_fourcc (structure, "format", &fourcc);
          vyp->fourcc[0] = (char)(fourcc         & 0xff);
          vyp->fourcc[1] = (char)((fourcc >> 8)  & 0xff);
          vyp->fourcc[2] = (char)((fourcc >> 16) & 0xff);
          vyp->fourcc[3] = (char)((fourcc >> 24) & 0xff);
          vyp->fourcc[4] = '\0';

          vp->type = EIC_VIDEO_RAW_YUV;
          vp->prop = vyp;
        }
      }
      else if (g_str_has_prefix (str, "video/x-raw-rgb")) {
        Eic_Video_Rgb_Prop *vrp;

        vrp = (Eic_Video_Rgb_Prop *)g_malloc (sizeof (Eic_Video_Rgb_Prop));
        if (vrp) {
          gst_structure_get_int (structure, "bpp", &vrp->bpp);
          gst_structure_get_int (structure, "depth", &vrp->depth);
          gst_structure_get_int (structure, "endianess", &vrp->endianess);
          gst_structure_get_int (structure, "red_mask", &vrp->red_mask);
          gst_structure_get_int (structure, "green_mask", &vrp->green_mask);
          gst_structure_get_int (structure, "blue_mask", &vrp->blue_mask);

          vp->type = EIC_VIDEO_RAW_RGB;
          vp->prop = vrp;
        }
      }
    }
    else if (g_str_has_prefix (str, "audio/")) {
      Eic_Audio_Prop *ap;
      GstStructure   *structure;

      ap = (Eic_Audio_Prop *)g_malloc (sizeof (Eic_Audio_Prop));
      if (!ap) continue;
      eic->audio_prop = g_list_append (eic->audio_prop, ap);
      ap->mimetype = mimetype;
      ap->decoder_info = decoder_info;
      ap->length_time = duration;
      ap->prop = NULL;

      structure = gst_caps_get_structure (GST_CAPS (caps), 0);
      gst_structure_get_int (structure, "rate", &ap->samplerate);
      gst_structure_get_int (structure, "channels", &ap->channels);
      ap->length_samples = ap->samplerate * ap->length_time / GST_SECOND;
      
      if (g_str_has_prefix (str, "audio/x-raw-int")) {
        Eic_Audio_Int_Prop *aip;
        
        aip = (Eic_Audio_Int_Prop *)g_malloc (sizeof (Eic_Audio_Int_Prop));
        if (aip) {
          gst_structure_get_int (structure, "endianess", &aip->endianess);
          gst_structure_get_int (structure, "width", &aip->width);
          gst_structure_get_int (structure, "depth", &aip->depth);
          gst_structure_get_boolean (structure, "signed_sampled", &aip->signed_samples);

          ap->type = EIC_AUDIO_RAW_INT;
          ap->prop = aip;
        }
      }
      else if (g_str_has_prefix (str, "audio/x-raw-float")) {
        Eic_Audio_Float_Prop *afp;
        
        afp = (Eic_Audio_Float_Prop *)g_malloc (sizeof (Eic_Audio_Float_Prop));
        if (afp) {
          gst_structure_get_int (structure, "endianess", &afp->endianess);
          gst_structure_get_int (structure, "width", &afp->width);
          gst_structure_get_int (structure, "buffer_frames", &afp->buffer_frames);

          ap->type = EIC_AUDIO_RAW_FLOAT;
          ap->prop = afp;
        }
      }
    }

    g_free (str);

    gst_object_unref (pad);
  }
  gst_iterator_free (it);

  res = gst_element_set_state (pipeline, GST_STATE_NULL);
  if (res == GST_STATE_CHANGE_ASYNC) {
    printf ("Could not go to NULL state\n");
    gst_object_unref (pipeline);
    return 0;
  }
  if (res == GST_STATE_CHANGE_FAILURE) {
    printf ("Could not go to NULL state\n");
    gst_object_unref (pipeline);
    return 0;
  }
  res = gst_element_get_state (pipeline, NULL, NULL, GST_CLOCK_TIME_NONE);
  if (res != GST_STATE_CHANGE_SUCCESS) {
    printf ("Failed to complete state change to NULL\n");
    gst_object_unref (pipeline);
    return 0;
  }

  return 1;
}

static void
new_decoded_pad_cb (GstElement *decodebin,
                    GstPad     *new_pad,
                    gboolean    last,
                    GstElement *pipeline)
{
  GstElement *fakesink;
  GstPad     *sinkpad;

  fakesink = gst_element_factory_make ("fakesink", NULL);
  gst_bin_add (GST_BIN (pipeline), fakesink);

  sinkpad = gst_element_get_pad (fakesink, "sink");
  if (GST_PAD_LINK_FAILED (gst_pad_link (new_pad, sinkpad))) {
    g_warning ("Failed to link %s:%s to %s:%s",
               GST_DEBUG_PAD_NAME (new_pad),
               GST_DEBUG_PAD_NAME (sinkpad));
    gst_bin_remove (GST_BIN (pipeline), fakesink);
  }
  else {
    gst_element_set_state (fakesink, GST_STATE_PAUSED);
  }
}

static void
element_added_cb (GstBin *bin,
                  GstElement *element,
                  gpointer user_data)
{
  Eic_Prop          *eic;
  GstElementFactory *factory;
  const gchar       *klass;

  eic = (Eic_Prop *)user_data;
  if (!eic)
    return;
  factory = gst_element_get_factory (element);
  klass = gst_element_factory_get_klass (factory);
  if (g_ascii_strcasecmp (klass, "Codec/Demuxer") == 0)
    eic->demuxer_info = eic_factory_new (factory);
}
